#ifndef MSP_GL_FONT_H_
#define MSP_GL_FONT_H_

#include <map>
#include <string>
#include <msp/datafile/objectloader.h>
#include <msp/stringcodec/utf8.h>

namespace Msp {
namespace GL {

class PrimitiveBuilder;
class Texture2D;

/**
Stores a set of glyphs and creates strings out of them.
*/
class Font
{
public:
	class Loader: public DataFile::CollectionObjectLoader<Font>
	{
	public:
		Loader(Font &, Collection &);

	private:
		void glyph(unsigned);
		void kerning(unsigned, unsigned, float);
		void ligature(unsigned, unsigned, unsigned);
		void texture();
		void texture_ref(const std::string &);
	};

	struct Glyph
	{
		class Loader: public Msp::DataFile::ObjectLoader<Glyph>
		{
		public:
			Loader(Glyph &);
		private:
			void texcoords(float, float, float, float);
		};

		unsigned code = 0;
		float x1 = 0.0f;
		float y1 = 0.0f;
		float x2 = 1.0f;
		float y2 = 1.0f;
		float w = 1.0f;
		float h = 1.0f;
		float off_x = 0.0f;
		float off_y = 0.0f;
		float advance = 1.0f;
	};

private:
	typedef std::map<unsigned, Glyph> GlyphMap;
	typedef std::pair<unsigned, unsigned> CodePair;
	typedef std::map<CodePair, float> KerningMap;
	typedef std::map<CodePair, unsigned> LigatureMap;

	const Texture2D *texture = 0;
	float native_size = 1.0f;
	float ascent = 1.0f;
	float descent = 0.0f;
	float cap_height = 1.0f;
	float x_height = 0.5f;
	GlyphMap glyphs;
	KerningMap kerning;
	LigatureMap ligatures;

public:
	void set_texture(const Texture2D &);
	const Texture2D &get_texture() const;

	/** Adds a glyph to the font.  There must not be an existing glyph with the
	same code. */
	void add_glyph(const Glyph &);
	void set_kerning(unsigned, unsigned, float);

	/** Returns the size used to generate the font texture.  This serves as a
	hint for obtaining the best quality when rendering strings. */
	float get_native_size() const { return native_size; }

	float get_ascent() const { return ascent; }
	float get_descent() const { return descent; }
	float get_cap_height() const { return cap_height; }
	float get_x_height() const { return x_height; }

	/** Returns the width of a string, in multiples of the font size.  Scale the
	result according to the size used in rendering. */
	float get_string_width(const std::string &, StringCodec::Decoder &) const;

	template<class C>
	float get_string_width(const std::string &str) const
	{
		typename C::Decoder dec;
		return get_string_width(str, dec);
	}

	float get_string_width(const std::string &str) const
	{ return get_string_width<StringCodec::Utf8>(str); }

	/** Builds the primitives for a string.  Two-dimensional vertex and texture
	coordinates are generated.  Size 1.0 is used for building; set up the
	builder's matrix before the call.  The texture is not bound, to avoid
	unnecessary bindings when creating meshes. */
	void build_string(const std::string &, StringCodec::Decoder &, PrimitiveBuilder &) const;

	template<class C>
	void build_string(const std::string &str, PrimitiveBuilder &pb) const
	{
		typename C::Decoder dec;
		build_string(str, dec, pb);
	}

	void build_string(const std::string &str, PrimitiveBuilder &pb) const
	{ return build_string<StringCodec::Utf8>(str, pb); }

private:
	void create_glyph_quad(const Glyph &, PrimitiveBuilder &) const;
	float get_glyph_advance(unsigned, unsigned = 0) const;
	unsigned get_ligature(unsigned, unsigned) const;
};

} // namespace GL
} // namespace Msp

#endif
