#include <msp/core/algorithm.h>
#include <msp/core/maputils.h>
#include <msp/strings/format.h>
#include "device.h"
#include "error.h"
#include "sequence.h"
#include "sequencebuilder.h"
#include "sequencetemplate.h"
#include "view.h"

using namespace std;

namespace Msp {
namespace GL {

SequenceBuilder::SequenceBuilder(const SequenceTemplate &t):
	tmpl(t)
{
	for(const SequenceTemplate::Renderable &r: tmpl.get_renderables())
		renderables[r.slot_name] = r.renderable;
	for(const SequenceTemplate::PostProcessor &p: tmpl.get_postprocessors())
		if(!p.slot_name.empty())
			postprocessors[p.slot_name] = 0;
}

void SequenceBuilder::set_renderable(const string &name, Renderable &rend)
{
	get_item(renderables, name) = &rend;
}

void SequenceBuilder::set_postprocessor(const string &name, PostProcessor &pproc)
{
	get_item(postprocessors, name) = &pproc;
}

void SequenceBuilder::set_debug_name(const string &name)
{
#ifdef DEBUG
	debug_name = name;
#else
	(void)name;
#endif
}

void SequenceBuilder::build(Sequence &sequence) const
{
#ifdef DEBUG
	if(!debug_name.empty())
		sequence.set_debug_name(debug_name);
#endif

	map<string, Renderable *> final_renderables = renderables;
	vector<const SequenceTemplate::Renderable *> sequences;
	for(const SequenceTemplate::Renderable &r: tmpl.get_renderables())
	{
		Renderable *&ptr = final_renderables[r.slot_name];
		if(!ptr)
		{
			if(r.effect_template)
			{
				Effect *effect = r.effect_template->create(final_renderables);
#ifdef DEBUG
				if(!debug_name.empty())
					effect->set_debug_name(format("%s/%s", debug_name, r.slot_name));
#endif
				ptr = effect;
			}
			else if(r.sequence_template)
			{
				Sequence *seq = new Sequence;
#ifdef DEBUG
				if(!debug_name.empty())
					seq->set_debug_name(format("%s/%s", debug_name, r.slot_name));
#endif
				ptr = seq;
				sequences.push_back(&r);
			}
			else
				throw invalid_operation("SequenceBuilder::build");

			sequence.add_owned(ptr);
		}
	}

	for(const SequenceTemplate::Renderable *s: sequences)
	{
		SequenceBuilder bld(*s->sequence_template);
		for(const auto &kvp: s->sequence_renderables)
			if(Renderable *r = get_item(final_renderables, kvp.second))
				bld.set_renderable(kvp.first, *r);
		bld.build(*static_cast<Sequence *>(get_item(final_renderables, s->slot_name)));
	}

	sequence.set_clear_enabled(tmpl.is_clear_enabled());
	if(tmpl.is_clear_enabled())
	{
		sequence.set_clear_colors(tmpl.get_clear_colors());
		sequence.set_clear_depth(tmpl.get_clear_depth());
		sequence.set_clear_stencil(tmpl.get_clear_stencil());
	}

	for(const SequenceTemplate::Step &s: tmpl.get_steps())
	{
		// Existence is already ensured above
		Renderable &renderable = *get_item(final_renderables, s.renderable_name);
		Sequence::Step &step = sequence.add_step(s.tag, renderable);
		step.set_depth_test(s.depth_test);
		step.set_stencil_test(s.stencil_test);
		step.set_lighting(s.lighting);
	}

#ifdef DEBUG
	unsigned index = 0;
#endif
	for(const SequenceTemplate::PostProcessor &p: tmpl.get_postprocessors())
	{
		PostProcessor *proc = 0;
		if(!p.slot_name.empty())
			proc = get_item(postprocessors, p.slot_name);
		if(!proc)
		{
			if(p.postprocessor_template)
			{
				proc = p.postprocessor_template->create(sequence.get_width(), sequence.get_height());
#ifdef DEBUG
				if(!debug_name.empty())
					proc->set_debug_name(format("%s/%d.pproc", debug_name, index++));
#endif
			}
			else
				throw invalid_operation("SequenceBuilder::build");

			sequence.add_owned(proc);
		}

		sequence.add_postprocessor(*proc);
	}
}

Sequence *SequenceBuilder::build() const
{
	RefPtr<Sequence> sequence = new Sequence();
	build(*sequence);
	return sequence.release();
}

Sequence *SequenceBuilder::build(unsigned w, unsigned h) const
{
	RefPtr<Sequence> sequence = create_sequence(w, h);
	build(*sequence);
	return sequence.release();
}

Sequence *SequenceBuilder::build(const View &view) const
{
	RefPtr<Sequence> sequence = create_sequence(view.get_width(), view.get_height());
	build(*sequence);
	return sequence.release();
}

Sequence *SequenceBuilder::build(const Framebuffer &fbo) const
{
	RefPtr<Sequence> sequence = create_sequence(fbo.get_width(), fbo.get_height());
	build(*sequence);
	return sequence.release();
}

Sequence * SequenceBuilder::create_sequence(unsigned w, unsigned h) const
{
	if(!w || !h || tmpl.get_postprocessors().empty())
		return new Sequence;

	unsigned samples = min(tmpl.get_maximum_multisample(), Device::get_current().get_info().limits.max_samples);
	if(samples<tmpl.get_required_multisample())
		throw invalid_operation("SequenceBuilder::create_frame_format");

	DataType color_type = (tmpl.get_hdr() ? HALF_FLOAT : UNSIGNED_BYTE);
	PixelFormat color_pf = make_pixelformat(RGBA, color_type);

	FrameFormat format = (COLOR_ATTACHMENT,color_pf, DEPTH_ATTACHMENT).set_samples(samples);
	return new Sequence(w, h, format);
}

} // namespace GL
} // namespace Msp
